package greenfoot;

import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.PointF;
import android.graphics.RectF;
import java.lang.Class;
import java.util.ArrayList;
import java.util.List;

/**
 * An Actor is an object that exists in the Greenfoot world. 
 * Every Actor has a location in the world, and an appearance (that is:
 * an icon).
 * 
 * <p>An Actor is not normally instantiated, but instead used as a superclass
 * to more specific objects in the world. Every object that is intended to appear
 * in the world must extend Actor. Subclasses can then define their own
 * appearance and behaviour.
 * 
 * <p>One of the most important aspects of this class is the 'act' method. This method
 * is called when the 'Act' or 'Run' buttons are activated in the Greenfoot interface.
 * The method here is empty, and subclasses normally provide their own implementations.
 * 
 * @author Upupzealot
 * @version 0.1
 */
public class Actor 
{
    World world;
    
    float x;
    float y;
    boolean rotated;
    float rotation;

    boolean scaled;
    float scaleX;
    float scaleY;
    GreenfootImage image;

    int paintOrder;
    int actOrder;

    /**
     * Construct an Actor.
     * The object would have a default image
     */
    public Actor(){
        x = 0;
        y = 0;
        rotated = false;
        rotation = 0;

        scaled = false;
        scaleX = 1;
        scaleY = 1;
        setImage(ClassImageManager.getClassImage(this.getClass()));

        paintOrder = 9999;
        actOrder = 9999;
    }

    /**
     * This method is called by the Greenfoot system when this actor has
     * been inserted into the world. This method can be overridden to implement
     * custom behaviour when the actor is inserted into the world.
     * <p>
     * The default implementation does nothing.
     * 
     * @param world The world the object was added to.
     */
    protected void addedToWorld(World world){
    }

    void act0(float dt){
        act();
    }
    
    /**
     * The act method is called by the greenfoot framework to give actors a
     * chance to perform some action. At each action step in the environment,
     * each object's act method is invoked, in unspecified order.
     * 
     * <p>The default implementation does nothing. This method should be overridden in
     * subclasses to implement an actor's action.
     */
    public void act(){
    }
    

    void draw(Canvas canvas){
        if(image == null){
            image = getClassImage();
        }
        else
            canvas.drawBitmap(image.bmp, 0, 0, image.paint);
        }
    
    void checkWorld(){
        if(world == null) {
            throw new IllegalStateException();
        }
    }
    
    /**
     * Return the world that this actor lives in.
     * @return The world.
     */
    public World getWorld(){
        checkWorld();
        return world;
    }
    
    /**
     * Sets the world, and the initial location. The location is adjusted according to the world's bounding
     * rules. The cached collision checking bounds, if any, are cleared.
     */
    //protected void addtoWorld(int px, int py, World world){
    //    world.addObject(this, x, y);
    //}

    
    void setX(int X){
        x = X;
    }
    void setX(float X){
        x = X;
    }

    /**
     * Return the x-coordinate of the actor's current location. The
     * value returned is the horizontal index of the actor's cell in the world.
     * 
     * @return The x-coordinate of the object's current location.
     * @throws IllegalStateException If the actor has not been added into a world.
     */
    public int getX(){
        checkWorld();
        return (int)x;
    }
    float getXF(){
		checkWorld();
        return x;
    }

    void setY(int Y){
        y = Y;
    }
    void setY(float Y){
        y = Y;
    }

    /**
     * Return the y-coordinate of the object's current location. The
     * value returned is the vertical index of the actor's cell in the world.
     * 
     * @return The y-coordinate of the actor's current location
     * @throws IllegalStateException If the actor has not been added into a world.
     */
    public int getY(){
        checkWorld();
        return (int)y;
    }
    float getYF(){
        return y;
    }
    
    /**
     * Assign a new location for this actor. This moves the actor to the specified
     * location. The location is specified as the coordinates of a cell in the world.
     * 
     * <p>If this method is overridden it is important to call this method as
     * "super.setLocation(x,y)" from the overriding method, to avoid infinite recursion.
     * 
     * @param X Location index on the x-axis
     * @param Y Location index on the y-axis
     * 
     * @see #move(int)
     */
    public void setLocation(int X, int Y){
        x = X;
        y = Y;
    }
    void setLocationF(float X, float Y){
        x = X;
        y = Y;
    }
    
    float getWidth(){
        return image.width / world.cellSize;
    }
    
    float getHeight(){
        return image.height / world.cellSize;
    }
    
    PointF getCornerPoint(float cx, float cy){
        float distance = distanceOf(getWidth() * cx, getHeight() * cy);
        double degrees = degreesBy(getWidth() * cx, getHeight() * cy);
        degrees += rotation;
        float px = x + distance * (float)Math.cos(Math.toRadians(degrees));
        float py = y + distance * (float)Math.sin(Math.toRadians(degrees));
        return new PointF(px, py);
    }
    
    PointF getLeftTop(){
        return getCornerPoint(-0.5f, -0.5f);
    }
    
    PointF getRightTop(){
        return getCornerPoint(0.5f, -0.5f);
    }
    
    PointF getLeftBottom(){
        return getCornerPoint(-0.5f, 0.5f);
    }

    PointF getRightBottom(){
        return getCornerPoint(0.5f, 0.5f);
    }
    
    double degreesBy(float dx, float dy){
        double degrees = Math.toDegrees(Math.atan2((double)dy, (double)dx));
        if(degrees < 0)
            degrees += 360;
        return degrees;
    }
    
    double degreesTo(float px, float py){
        float dx = px - x;
        float dy = py - y;
        return degreesBy(dx, dy);
    }
    
    boolean isContain(float px, float py){
        float distance = distanceTo(px, py);
        double degrees = degreesTo(px, py);
        
        float dx = distance * (float)Math.cos(Math.toRadians(degrees - (double)rotation));
        float dy = distance * (float)Math.sin(Math.toRadians(degrees - (double)rotation));
        
        RectF rect = new RectF(-getWidth() / 2, -getHeight() / 2, getWidth() / 2, getHeight() / 2);
        return rect.contains(dx, dy);
    }
	boolean isContain(PointF point){
		return isContain(point.x, point.y);
	}
    
    /**
     * Check whether this object intersects with another given object.
     * 
     * @return True if the object's intersect, false otherwise.
     */
    protected boolean intersects(Actor actor){
      return (isContain(actor.getLeftTop())
	       || isContain(actor.getLeftBottom())
		   || isContain(actor.getRightTop())
		   || isContain(actor.getRightBottom())
		   || actor.isContain(getLeftTop())
		   || actor.isContain(getLeftBottom())
		   || actor.isContain(getRightTop())
		   || actor.isContain(getRightBottom())
		   );
    }
	
    
    /**
     * Move this actor the specified distance in the direction it is currently facing.<br>
     * The direction can be set using the setRotation(int) method.
     * param distance The distance to move (in cell-size units); a negative value will move backwards
     * @see #setLocation(int, int)
     */
    public void move(int distance){
        moveF(distance);
    }
    void moveF(float distance){
        double theta = Math.toRadians(rotation);
        x += distance * Math.cos(theta);
        y += distance * Math.sin(theta);
    }
    
    float distanceOf(float dx, float dy){
        return (float)Math.sqrt((double)(dx * dx + dy * dy));
    }
    
    float distanceOf(float x1, float y1, float x2, float y2){
        float dx = x2 - x1;
        float dy = y2 - y1;
        return distanceOf(dx, dy);
    }
    
    float distanceTo(float px, float py){
        return distanceOf(x, y, px, py);
    }
    
    float distanceTo(Actor actor){
        return distanceTo(actor.x, actor.y);
    }

    /**
     * Set the rotation of this actor. Rotation is expressed as a degree value, range (0..359). Zero degrees is to the east (right-hand side of the world), and the angle increases clockwise.
     * @param r The rotation in degrees.
     * @see #turn(int)
     */
    public void setRotation(int r){
        setRotationF(r);
    }
    void setRotationF(float r){
        rotation = r;
        if(rotation != 0)
            rotated = true;
        else
            rotated = false;
    }
    
    /**
     * Turn this actor to face towards a certain location.
     * 
     * @param px  The x-coordinate of the cell to turn towards
     * @param py  The y-coordinate of the cell to turn towards
     */
    public void turnTowards(int px, int py){
        turnTowardsF(px, py);
    }
    void turnTowardsF(float px, float py){
        setRotationF((float)degreesTo(px, py));
    }

    
    /**
     * Turn this actor by the specified amount (in degrees).
     * @param amount the number of degrees to turn; positive values turn clockwise
     * @see #setRotation(int)
     */
    public void turn(int amount){
        turnF(amount);
    }
    void turnF(float amount){
        rotation += amount;
    }

    /**
     * Return the current rotation of this actor. Rotation is expressed as a degree value, range (0..359). Zero degrees is towards the east (right-hand side of the world), and the angle increases clockwise.
     * @return The rotation in degrees.
     * @see #setRotation
     */
    public int getRotation(){
        int r = Math.round(getRotationF());
        r = r % 360;
        return r;
    }
    float getRotationF(){
        return rotation;
    }
    

    /**
     * <font color = #8600FF>
     * Returns the image used to represent this Class. This image can be
     * modified to change all the actors' appearance of this Class which were not generated yet.
     * 
     * @return The Class's image.
     * </font>
     */
    public static GreenfootImage getClassImage(){
        return ClassImageManager.getClassImage(
        new Object(){
            public Class getCurrentClass(){
                return this.getClass();
            }
        }.getCurrentClass()
        );
    }

    /**
     * Set the image for this actor to the specified image.
     * 
     * @see #setImage(String)
     * @param image The image.
     */
    public void setImage(GreenfootImage i){
        image = i;
    }
    
    /**
     * Set an image for this actor from an image file. The file may be in
     * jpeg, <font color = #8600FF><s>gif</s></font> or png format. The file should be located in the project
     * directory.
     * 
     * @param filename The name of the image file.
     * @throws IllegalArgumentException If the image can not be loaded.
     */
    public void setImage(String filename){
        GreenfootImage zi = new GreenfootImage(filename);
        setImage(zi);
    }

    /**
     * Returns the image used to represent this actor. This image can be modified to change the actor's appearance.
     * @return The object's image.
     */
    public GreenfootImage getImage(){
        return image;
    }
    
    boolean isChildOf(Class cls){
        return cls.isAssignableFrom(getClass());
    }
    
    
    
    
    /**
     * Return the neighbours to this object within a given distance. This method considers only logical location, ignoring extent of the image. Thus, it is most useful in scenarios where objects are contained in a single cell.<br>
     * All cells that can be reached in the number of steps given in 'distance' from this object are considered. Steps may be only in the four main directions, or may include diagonal steps, depending on the 'diagonal' parameter. Thus, a distance/diagonal specification of (1,false) will inspect four cells, (1,true) will inspect eight cells. 
     * @param distance Distance (in cells) in which to look for other objects.
     * @param diagonal If true, include diagonal steps.
     * @param cls Class of objects to look for (passing 'null' will find all objects). 
     * @return A list of all neighbours found.
     */
    protected List getNeighbours(float distance, boolean diagonal, Class cls){
		checkWorld();
        List<Actor> aimList= new ArrayList<Actor>();
        List actors = getObjectsInRange(distance, cls);
        for(int i = 0; i < actors.size(); i++){
            Actor actor = (Actor)actors.get(i);
            if(diagonal)
                aimList.add(actor);
            else
            if(getX() == actor.getX() && getY() == actor.getY())
                aimList.add(actor);
        }
        return aimList;
    }
    
    /**
     * Return all objects that intersect the center of the given location (relative to this object's location). 
     * @param dx X-coordinate relative to this objects location.
     * @param dy y-coordinate relative to this objects location.
     * @param cls Class of objects to look for (passing 'null' will find all objects). 
     * @return List of objects at the given offset. The list will include this object, if the offset is zero.
     */
    protected List getObjectsAtOffset(int dx, int dy, Class cls){
        checkWorld();
        List<Actor> aimList= new ArrayList<Actor>();
        List actors = world.getObjects(cls);
        for(int i = 0; i < actors.size(); i++){
            Actor actor = (Actor)actors.get(i);
            if(actor.getX() - getX() == dx && actor.getY() - getY() == dy)
                aimList.add(actor);
        }
		aimList.remove(this);
        return aimList;
    }
    
    /**
     * Return one object that is located at the specified cell (relative to this objects location). Objects found can be restricted to a specific class (and its subclasses) by supplying the 'cls' parameter. If more than one object of the specified class resides at that location, one of them will be chosen and returned.
     * @param dx X-coordinate relative to this objects location.
     * @param dy y-coordinate relative to this objects location.
     * @param cls Class of objects to look for (passing 'null' will find all objects). 
     * @return An object at the given location, or null if none found.
     */
    protected Actor getOneObjectAtOffset(int dx, int dy, Class cls){
        if(getObjectsAtOffset(dx, dy, cls).isEmpty())
            return null;
        else
            return (Actor)getObjectsAtOffset(dx, dy, cls).get(0);
    }
    
    /**
     * Return all objects within range 'radius' around this object. An object is within range if the distance between its centre and this object's centre is less than or equal to 'radius'.
     * @param radius Radius of the circle (in cells)
     * @param cls - Class of objects to look for (passing 'null' will find all objects).
     */
    protected List getObjectsInRange(float radius, Class cls){
		checkWorld();
        List<Actor> neighbourList= new ArrayList<Actor>();
        List actors = world.getObjects(cls);
        for(int i = 0; i < actors.size(); i++){
            Actor actor = (Actor)actors.get(i);
            if(distanceTo(actor) <= radius * getWorld().cellSize)
            neighbourList.add(actor);
        }
		neighbourList.remove(this);
        return neighbourList;
    }
    
	
	/**
     * Return all the objects that intersect this object. This takes the graphical extent of objects into consideration. 
     * @param cls Class of objects to look for (passing 'null' will find all objects).
     */
    protected List getIntersectingObjects(Class cls){
		checkWorld();
		List<Actor> aimList= new ArrayList<Actor>();
        List actors = world.getObjects(cls);
        for(int i = 0; i < actors.size(); i++){
            Actor actor = (Actor)actors.get(i);
            if(intersects(actor))
                aimList.add(actor);
        }
		aimList.remove(this);
        return aimList;
    }
	
	/**
     * Return an object that intersects this object. This takes the graphical extent of objects into consideration. 
     * @param cls Class of objects to look for (passing 'null' will find all objects).
     */
    protected Actor getOneIntersectingObject(Class cls){
        if(getIntersectingObjects(cls).isEmpty())
            return null;
        else
            return (Actor)getIntersectingObjects(cls).get(0);
    }
}
