package greenfoot;

import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.PorterDuff.Mode;
import android.graphics.PorterDuffXfermode;
import java.io.InputStream;
import java.io.IOException;

public final class GreenfootImage
{
    Bitmap bmp;
    int width;
    int height;

    Canvas canvas;
    Paint paint;

    GreenfootImage(){
        new GreenfootImage(1, 1);
    }

    /**
     * Create an empty (transparent) image with the specified size.
     * 
     * @param w The width of the image in pixels.
     * @param h The height of the image in pixels.
     */
    public GreenfootImage(int w, int h){
        bmp = Bitmap.createBitmap(w, h, Bitmap.Config.ARGB_8888);
        init();
    }

    /**
     * Create an image from an image file. Supported file formats are JPEG, GIF
     * and PNG.
     * 
     * <p><The file name may be <s>an absolute path</s>, or a base name for a file located in
     * the project directory.
     * 
     * @param filename Typically the name of a file in the images directory within
     *            the project directory.
     * @throws IllegalArgumentException If the image can not be loaded.
     */
    public GreenfootImage(String fileName){
        try{
            InputStream is = Greenfoot.activity.getAssets().open("images/" + fileName);
            bmp = BitmapFactory.decodeStream(is);
            is.close();
			Bitmap b = bmp.copy(bmp.getConfig(), true);
			bmp.recycle();
			bmp = b;
			
            init();
        }
		catch(IOException ee){
            throw new java.lang.IllegalArgumentException("  loadind file  \"" + fileName + "\"  faild");
        }
    }

    /**
     * Create a GreenfootImage from another GreenfootImage.
     */
    public GreenfootImage(GreenfootImage image){
        bmp = Bitmap.createBitmap(image.bmp);
        init();
    }

    /**
     * Creates an image with the given string drawn as text using the given font size, with the given foreground
     * color on the given background color.  If the string has newline characters, it
     * is split into multiple lines which are drawn horizontally-centred.
     * 
     * @param string the string to be drawn
     * @param size the requested height in pixels of each line of text (the actual height may be different by a pixel or so)
     * @param foreground the color of the text.  Since Greenfoot 2.2.0, passing null will use black.
     * @param background the color of the image behind the text.  Since Greenfoot 2.2.0, passing null with leave the background transparent.
     * @since 2.0.1
     */
    public GreenfootImage(String string, int size, Color foreground, Color background){
		Rect rect = new Rect();
		paint = new Paint();
		paint.setTextSize(size);
		paint.getTextBounds(string, 0, string.length(), rect);
		bmp = Bitmap.createBitmap(rect.width() + 4, rect.height() + 4, Bitmap.Config.ARGB_8888);
		init();
		paint.setTextSize(size);
		setColor(background);
		fill();
		setColor(foreground);
		drawString(string, 2, rect.height() + 2);
    }

    void init(){
        width = bmp.getWidth();
        height = bmp.getHeight();

        canvas = new Canvas(bmp);
        
        if(paint == null){
			paint = new Paint();
			paint.setColor(android.graphics.Color.BLACK);
			paint.setAlpha(255);
			//paint.setAntiAlias(true);
		}
    }

    GreenfootImage(GreenfootImage image, int left, int top, int width, int height){
        bmp = Bitmap.createBitmap(image.bmp, left, top, width, height);
		init();
    }

    /**
     * Return the width of the image.
     * 
     * @return Width of the image.
     */

    public int getWidth(){
        return width;
    }

    /**
     * Return the height of the image.
     * 
     * @return Height of the image.
     */
    public int getHeight(){
        return height;
    }
    
    void checkBounds(float x, float y){
        if (x >= getWidth()) {
            throw new IndexOutOfBoundsException("X is out of bounds. It was: " + x
                    + " and it should have been smaller than: " + getWidth());
        }
        if (y >= getHeight()) {
            throw new IndexOutOfBoundsException("Y is out of bounds. It was: " + y
                    + " and it should have been smaller than: " + getHeight());
        }
        if (x < 0) {
            throw new IndexOutOfBoundsException("X is out of bounds. It was: " + x
                    + " and it should have been at least: 0");
        }
        if (y < 0) {
            throw new IndexOutOfBoundsException("Y is out of bounds. It was: " + y
                    + " and it should have been at least: 0");
        }
    }

    /**
     * Set the current drawing color. This color will be used for subsequent
     * drawing operations.
     * 
     * @param color The color to be used.
     */
    public void setColor(Color color){
        paint.setColor(color.getAndroidColor());
    }

    /**
     * Sets the color at the given pixel to the given color.
     */
    public void setColorAt(int x, int y, Color color){
        checkBounds(x, y);
        bmp.setPixel(x, y, color.getAndroidColor());
    }
    
    void setColor(int r, int g, int b, int a){
        paint.setColor(new Color(r, g, b, a).getAndroidColor());
    }

    /**
     * Clears the image.
     */
    public void clear(){
        //canvas.drawARGB(0, 255, 255, 255);
		Paint p = new Paint();
		p.setXfermode(new PorterDuffXfermode(Mode.CLEAR));
		canvas.drawRect(0, 0, width, height, p);
    }

    /**
     * Fill the entire image with the current drawing dcolor.
     * 
     */
    public void fill(){
        canvas.drawColor(paint.getColor());
    }

    /**
     * Draws the given Image onto this image
     * @param i The image to draw onto this one.
     * @param x x-coordinate for drawing the image.
     * @param y y-coordinate for drawing the image.
     */
	public void drawImage(GreenfootImage i, int x, int y){
		drawImageF(i, x, y);
	}
	void drawImageF(GreenfootImage i, float x, float y){
		canvas.drawBitmap(i.bmp, x, y, i.paint);
	}

	/**
     * Draw a line, using the current drawing color, between the points (x1, y1) and (x2, y2).
     * @param x1 the first point's x coordinate.
     * @param y1 the first point's y coordinate.
     * @param x2 the second point's x coordinate.
     * @param y2 the second point's y coordinate.
     */
	public void drawLine(int x1, int y1, int x2, int y2){
		drawLineF(x1, y1, x2, y2);
	}
    void drawLineF(float x1, float y1, float x2, float y2){
        canvas.drawLine(x1, y1, x2, y2, paint);
    }

    /**
     * Draw an oval bounded by the specified rectangle with the current drawing
     * color.
     * 
     * @param x the <i>x </i> coordinate of the upper left corner of the oval to
     *            be filled.
     * @param y the <i>y </i> coordinate of the upper left corner of the oval to
     *            be filled.
     * @param w the width of the oval to be filled.
     * @param h the height of the oval to be filled.
     */
	public void drawOval(int x, int y, int w, int h){
		drawOvalF(x, y, w, h);
	}
    void drawOvalF(float x, float y, float w, float h){
        paint.setStyle(Paint.Style.STROKE);
        RectF rect = new RectF(x, y, x + w, y + h);
        canvas.drawOval(rect, paint);
    }

    /**
     * Fill an oval bounded by the specified rectangle with the current drawing
     * color.
     * 
     * @param x the <i>x </i> coordinate of the upper left corner of the oval to
     *            be filled.
     * @param y the <i>y </i> coordinate of the upper left corner of the oval to
     *            be filled.
     * @param w the width of the oval to be filled.
     * @param h the height of the oval to be filled.
     */
	public void fillOval(int x, int y, int w, int h){
		fillOvalF(x, y, w, h);
	}
    void fillOvalF(float x, float y, float w, float h){
        paint.setStyle(Paint.Style.FILL);
        RectF rect = new RectF(x, y, x + w, y + h);
        canvas.drawOval(rect, paint);
    }

    /**
     * Draws a closed polygon defined by arrays of <i>x</i> and <i>y</i>
     * coordinates. Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates
     * defines a point.
     * <p>
     * This method draws the polygon defined by <code>nPoint</code> line
     * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code> line
     * segments are line segments from
     * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code>
     * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for 1&nbsp;&le;&nbsp;<i>i</i>&nbsp;&le;&nbsp;<code>nPoints</code>.
     * The figure is automatically closed by drawing a line connecting the final
     * point to the first point, if those points are different.
     * 
     * @param xPoints an array of <code>x</code> coordinates.
     * @param yPoints an array of <code>y</code> coordinates.
     * @param nPoints the total number of points.
     */
	public void drawPolygon(int[] xPoints, int[] yPoints, int nPoints){
		paint.setStyle(Paint.Style.STROKE);
        Path path = new Path();
        path.moveTo(xPoints[0], yPoints[0]);
        for(int i = 1; i < nPoints; i++)
            path.moveTo(xPoints[i], yPoints[i]);
        path.close();
        canvas.drawPath(path, paint);
	}
    void drawPolygonF(float[] xPoints, float[] yPoints, int nPoints){
        paint.setStyle(Paint.Style.STROKE);
        Path path = new Path();
        path.moveTo(xPoints[0], yPoints[0]);
        for(int i = 1; i < nPoints; i++)
            path.moveTo(xPoints[i], yPoints[i]);
        path.close();
        canvas.drawPath(path, paint);
    }

    /**
     * Fill a closed polygon defined by arrays of <i>x </i> and <i>y </i>
     * coordinates.
     * <p>
     * This method draws the polygon defined by <code>nPoint</code> line
     * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code> line
     * segments are line segments from
     * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code>
     * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for 1&nbsp;&le;&nbsp;
     * <i>i </i>&nbsp;&le;&nbsp; <code>nPoints</code>. The figure is
     * automatically closed by drawing a line connecting the final point to the
     * first point, if those points are different.
     * <p>
     * The area inside the polygon is defined using an even-odd fill rule, also
     * known as the alternating rule.
     * 
     * @param xPoints a an array of <code>x</code> coordinates.
     * @param yPoints a an array of <code>y</code> coordinates.
     * @param nPoints a the total number of points.
     */
	public void fillPolygon(int[] xPoints, int[] yPoints, int nPoints){
		paint.setStyle(Paint.Style.FILL);
        Path path = new Path();
        path.moveTo(xPoints[0], yPoints[0]);
        for(int i = 1; i < nPoints; i++)
            path.lineTo(xPoints[i], yPoints[i]);
        path.close();
        canvas.drawPath(path, paint);
	}
    public void fillPolygonF(float[] xPoints, float[] yPoints, int nPoints){
        paint.setStyle(Paint.Style.FILL);
        Path path = new Path();
        path.moveTo(xPoints[0], yPoints[0]);
        for(int i = 1; i < nPoints; i++)
            path.lineTo(xPoints[i], yPoints[i]);
        path.close();
        canvas.drawPath(path, paint);
    }

    /**
     * Draw the outline of the specified rectangle. The left and right edges of
     * the rectangle are at <code>x</code> and
     * <code>x&nbsp;+&nbsp;width</code>. The top and bottom edges are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height</code>. The rectangle
     * is drawn using the current color.
     * 
     * @param x the <i>x </i> coordinate of the rectangle to be drawn.
     * @param y the <i>y </i> coordinate of the rectangle to be drawn.
     * @param w the width of the rectangle to be drawn.
     * @param h the height of the rectangle to be drawn.
     */
	public void drawRect(int x, int y, int w, int h){
		drawRectF(x, y, w, h);
	}
    void drawRectF(float x, float y, float w, float h){
        paint.setStyle(Paint.Style.STROKE);
        RectF rect = new RectF(x, y, x + w, y + h);
        canvas.drawRect(rect, paint);
    }

    /**
     * Fill the specified rectangle. The left and right edges of the rectangle
     * are at <code>x</code> and
     * <code>x&nbsp;+&nbsp;width&nbsp;-&nbsp;1</code>. The top and bottom
     * edges are at <code>y</code> and
     * <code>y&nbsp;+&nbsp;height&nbsp;-&nbsp;1</code>. The resulting
     * rectangle covers an area <code>width</code> pixels wide by
     * <code>height</code> pixels tall. The rectangle is filled using the
     * current color.
     * 
     * @param x the <i>x </i> coordinate of the rectangle to be filled.
     * @param y the <i>y </i> coordinate of the rectangle to be filled.
     * @param w the width of the rectangle to be filled.
     * @param h the height of the rectangle to be filled.
     */
	public void fillRect(int x, int y, int w, int h){
		fillRectF(x, y, w, h);
	}
    void fillRectF(float x, float y, float w, float h){
        paint.setStyle(Paint.Style.FILL);
        RectF rect = new RectF(x, y, x + w, y + h);
        canvas.drawRect(rect, paint);
    }

    /**
     * Draw the text given by the specified string, using the current font and
     * color. The baseline of the leftmost character is at position ( <i>x
     * </i>,&nbsp; <i>y </i>).
     * 
     * @param string the string to be drawn.
     * @param x the <i>x </i> coordinate.
     * @param y the <i>y </i> coordinate.
     */
	public void drawString(String string, int x, int y){
		drawStringF(string, x, y);
	}
    void drawStringF(String string, float x, float y){
        canvas.drawText(string, x, y, paint);
    }

    /**
     * Returns the android.graphics.Bitmap that backs this GreenfootImage. Any changes to
     * the returned image will be reflected in the GreenfootImage.
     * 
     * @return The android.graphics.Bitmap backing this GreenfootImage
     * @since Greenfoot version 0.1
     */
    public Bitmap getAndroidImage(){
        return bmp;
    }

    /**
     * Return the current drawing color.
     * @return The current color.
     */
    public Color getColor(){
        return new Color(paint.getColor());
    }

    /**
     * Return the color at the given pixel.
     * @throws java.lang.IndexOutOfBoundsException - If the pixel location is not within the image bounds.
     */
    public Color getColorAt(int x, int y){
        checkBounds(x, y);
        return new Color(bmp.getPixel(x, y));
    }

    /**
     * Return the current transparency of the image.
     * @return A value in the range 0 to 255. 0 is completely transparent (invisible) and 255 is completely opaque (the default).
     */
    public int getTransparency(){
        return paint.getAlpha();
    }

    /**
     * Set the transparency of the image.
     * param t A value in the range 0 to 255. 0 is completely transparent (invisible) and 255 is completely opaque (the default).
     */
    public void setTransparency(int t){
        check255(t, "transparency value");
        paint.setAlpha(t);
    }
    
    void check255(int n, String name){
        if (n < 0 || n > 255) {
            throw new IllegalArgumentException("The" + name + "has to be in the range 0 to 255. It was: " + n);
        }
    }
    
    /**
     * Get the current font.
     */
    public Font getFont(){
		return new Font("", paint.getTypeface().getStyle(), (int)paint.getTextSize());
	}
    
	/**
     * Set the current font. This font will be used for subsequent text
     * operations.
     */
	public void setFont(Font font){
		paint.setTypeface(font.tf);
		paint.setTextSize(font.size);
	}

    /**
     * Scales this image to a new size.
     * 
     * @param w Width of new image
     * @param h Height of new image
     */
    public void scale(int w, int h){
        Bitmap b = Bitmap.createScaledBitmap(bmp, w, h, false);
        bmp.recycle();
        bmp = b;
        init();
    }

    /**
     * Rotates this image around the center.
     * 
     * @param degrees degrees to rotate
     */
	public void rotate(int degrees){
		rotateF(degrees);
	}
    void rotateF(float degrees){
        Bitmap b =Bitmap.createBitmap(bmp.getWidth(), bmp.getHeight(), bmp.getConfig());
        Canvas c = new Canvas(b);
        c.save();
        c.rotate(degrees, bmp.getWidth() / 2f, bmp.getHeight() / 2f);
        c.drawBitmap(bmp, 0, 0, new Paint());
        c.restore();
        bmp.recycle();
        bmp = b;
        init();
    }

    /**
     * Mirrors the image vertically (the top of the image becomes the bottom, and vice versa).
     * 
     */
    public void mirrorHorizontally(){
        Bitmap b =Bitmap.createBitmap(bmp.getWidth(), bmp.getHeight(), bmp.getConfig());
        Canvas c = new Canvas(b);
        c.save();
        c.scale(1, -1, bmp.getWidth() / 2f, bmp.getHeight() / 2f);
        c.drawBitmap(bmp, 0, 0, new Paint());
        c.restore();
        bmp.recycle();
        bmp = b;
        init();
    }

    /**
     * Mirrors the image horizontally (the left of the image becomes the right, and vice versa).
     * 
     */
    public void mirrorVertically(){
        Bitmap b =Bitmap.createBitmap(bmp.getWidth(), bmp.getHeight(), bmp.getConfig());
        Canvas c = new Canvas(b);
        c.save();
        c.scale(-1, 1, bmp.getWidth() / 2f, bmp.getHeight() / 2f);
        c.drawBitmap(bmp, 0, 0, new Paint());
        c.restore();
        bmp.recycle();
        bmp = b;
        init();
    }
    
    /**
     * Return a text representation of the image.
     */
    public String toString()
    {
        return super.toString();
    }
}
